################# Function for generating cross validated predictions using MLR ##################
#
# Created by Pablo Mendoza, March 8 2015
#
##################################################################################################
#
# This function generates cross-validated predictions using multiple linear regression (MLR).
# The best model is selected based on the minimum RMSE.
#
# Inputs required:
#
# Xpredictor  :  Matrix containing predictors to be used for training ( [Nsyears x (Ngroups-1)] x Npred )
# Xpredict    :  Matrix to be used for predicting ( Nsyears x Npred )
# Ypredictand :  Column vector containing the predictors to be used ( [Nsyears x (Ngroups-1)] x1 )
# combinations:  Logical (TRUE/FALSE) matrix containing all predictor combinations to be considered
# Nsyears     :  Number of years left out for cross-validation
#
##################################################################################################


PredictKernelCrossVal <- function(Xpredictor,Xpredict,Ypredictand,combinations,Nsyears) {
  
  # Load required libraries
  library(np)
  
  if( length(combinations) == 1){ # In case we have only one predictor
    Xpred       = Xpredictor
    Xfor        = Xpredict
  }
  
  if( length(combinations) > 1){  # For all the rest of the cases...
    Ncomb       = dim(combinations)[1]       # Number of combinations of predictors
    output_comb = mat.or.vec(Ncomb,1)        # Output containing RMSE values
    Ntime = length(Ypredictand)
    
    # Start loop over combinations of predictors
    for (icomb in 1:Ncomb){
      
      Xpredsub  = Xpredictor[,combinations[icomb,]]
      kermod    = npreg( tydat=as.vector(Ypredictand), txdat=Xpredsub, regtype="ll", bwmethod="cv.ls")
      output_comb[icomb]  = kermod$MSE
      
    } # End loop over the combination of predictors
    
    where       = min(which(output_comb == min(output_comb))) # Location of the best set of predictors
    Xpred       = Xpredictor[,combinations[where,]]      # Best set of predictors
    if( Nsyears == 1 ){ # If only one year is left out
      Xfor      = Xpredict[combinations[where,]]         # Predictors for year iyear
    }
    if( Nsyears >  1 ){ # If more than 1 year is left out
      Xfor      = Xpredict[,combinations[where,]]        # Predictors for year iyear
    }
    
  } # End loop over combinations of predictors
  
  
  # Fit MLR model for year iyear and predict!
  X           = Xpred
  
  # Case # 1: Only one year dropped out (i.e. leave-one-out cross validation)
  if( Nsyears ==1 ){
    bestmodel   = npreg( tydat=as.vector(Ypredictand), txdat=X, exdat = t(Xfor),
                         regtype="ll", bwmethod="cv.ls",se.fit=TRUE)
    Yhat        = bestmodel$merr
    Yse         = bestmodel$merr
  }
  
  # Case # 2: More than one year dropped out (i.e. leave-one-out cross validation)
  if( Nsyears > 1 ){
    bestmodel   = npreg( tydat=as.vector(Ypredictand), txdat=X, exdat = Xfor,
                         regtype="ll", bwmethod="cv.ls",se.fit=TRUE)
    Yhat        = bestmodel$merr
    Yse         = bestmodel$merr
  }
  
  # Generate vector
  if( length(combinations) == 1 ){
    bestcomb='TRUE'
  }
  if( length(combinations)  > 1 ){
    bestcomb=combinations[where,]
  }
  
  Yreturn     = list(Yhat=Yhat,Yse=Yse,bestcomb=bestcomb)
  return(Yreturn)
  
}
