########## Function for generating cross validated predictions using local polynomials ###########
#
# Created by Pablo Mendoza, March 8 2015
#
##################################################################################################
#
# This function generates cross-validated ensemble predictions using local polynomials.
# The best model is selected based on the minimum generalized cross validation score (GCV).
# The reader is referred to Loader (1999) for more details on local regression methods.
# Similar applications can be found in Grantz et al. (2005), Regonda et al. (2006) and
# Bracken et al. (2010).
#
# Inputs required:
#
# min_alph    :  Minimum alpha to be used
# Xpredictor  :  Matrix containing predictors to be used for training (Nyears-1 x Npred)
# Xpredict    :  Row vector to be used for predicting
# Ypredictand :  Column vector containing the predictors to be used (Nyears-1 x1)
# combinations:  Logical (TRUE/FALSE) matrix containing all predictor combinations to be considered
# Nens        :  Number of ensemble members
#
##################################################################################################


PredictLocPolyCrossVal <- function(min_alph,Xpredictor,Xpredict,Ypredictand,combinations,Nens) {
  
  # Load required libraries
  library(locfit)
  
  # Set parameters for local polynomial
  alpha_grid  = seq(min_alph,1.0,by=0.05)
  n           = length(alpha_grid)
  alpha1_grid = seq(min_alph,1.0,by=0.05)
  alpha2_grid = c(alpha_grid,alpha1_grid)
  
  Ncomb       = dim(combinations)[1]       # Number of combinations of predictors
  output_comb = mat.or.vec(Ncomb,3)        # Output containing statistics from each combination
  
  # Start loop over combinations of predictors
  for (icomb in 1:Ncomb){
    
    Xpredsub  = Xpredictor[,combinations[icomb,]]
    gcv_deg1  = gcvplot(Ypredictand ~ ., data=as.data.frame(Xpredsub),maxk=10000000,maxit=100,alpha=alpha_grid,deg=1,kern="bisq", scale = T, family="gaussian",ev=dat())
    gcv_deg2  = gcvplot(Ypredictand ~ ., data=as.data.frame(Xpredsub),maxk=10000000,maxit=100,alpha=alpha1_grid,deg=2,kern="bisq", scale = T, family="gaussian",ev=dat())
    # pick the best alpha and the degree of the polynomial that
    # gives the smallest GCV
    z2                   = order(c(gcv_deg1$values,gcv_deg2$values))
    deg1                 = 1
    if(z2[1] > n) { deg1=2 }
    best_alpha           = alpha2_grid[z2[1]]
    best_gcv             = c(gcv_deg1$values,gcv_deg2$values)[z2[1]]
    output_comb[icomb,]  = t(c(deg1, best_alpha, best_gcv))    # save scores for this combination of predictors
    
  } # End loop over the combination of predictors
   
  where       = which((output_comb[,3]) == min(output_comb[,3])) # Location of the best set of predictors
  Xpred       = Xpredictor[,combinations[where,]]                # Best set of predictors
  Xfor        = Xpredict[combinations[where,]]                   # Predictors for year iyear
  bestdeg     = output_comb[where,1]                             # Best degree of polynomial
  best_alpha  = output_comb[where,2]                             # Best alpha
  
  # Fit the local polynomial for year iyear and predict!
  bestmodel   = locfit(Ypredictand ~ Xpred, maxk = 10000,maxit=100,alpha=best_alpha,deg=bestdeg,kern="bisq",scale =T,family="gaussian")
  zp          = predict(bestmodel,t(Xfor),se.fit=T,type='response')
  Yhat        = zp$fit
  Yse         = zp$se
  
  # Generate ensemble
  Yens        = rep(Yhat,Nens)+rnorm(Nens,0,sd=Yse)
  return(Yens)
  
}