################# Function for generating cross validated predictions using MLR ##################
#
# Created by Pablo Mendoza, March 8 2015
#
##################################################################################################
#
# This function generates cross-validated predictions using multiple linear regression (MLR).
# The best model is selected based on the minimum RMSE.
#
# Inputs required:
#
# Xpredictor  :  Matrix containing predictors to be used for training ( [Nsyears x (Ngroups-1)] x Npred )
# Xpredict    :  Matrix to be used for predicting ( Nsyears x Npred )
# Ypredictand :  Column vector containing the predictors to be used ( [Nsyears x (Ngroups-1)] x1 )
# combinations:  Logical (TRUE/FALSE) matrix containing all predictor combinations to be considered
# Nsyears     :  Number of years left out for cross-validation
#
##################################################################################################


PredictMLRCrossValv2 <- function(Xpredictor,Xpredict,Ypredictand,Nsyears,maxp) {
  
  # Load required libraries
  library(locfit)   # Library used to compute Akaike's Information Criterion
  library(relaimpo) # Library to compute relative importante of predictors
  
  Npred    = ifelse(is.null(ncol(Xpredictor)),1,ncol(Xpredictor))
  
  if( Npred == 1){ # In case we have only one predictor
    Xpred       = Xpredictor
    Xfor        = Xpredict
    Npredf      = 1
  }
  
  if( Npred > 1){  # For all the rest of the cases...
    
    corpred   = sapply(1:Npred, function(ipred) (cor(Xpredictor[,ipred],Ypredictand))^2)
    selpred   = which(corpred==max(corpred))
    Npredf    = length(selpred)
    mlrmod    = lm(Ypredictand ~ Xpredictor[,selpred])
    maxr2     = summary(mlrmod)$r.squared
    
    while(Npredf < maxp){
      corpred2  = sapply(setdiff(1:Npred,selpred), function(ipred) (cor(Xpredictor[,ipred],mlrmod$residuals))^2 )
      candidate = setdiff(1:Npred,selpred)[sort(corpred2,decreasing=TRUE,index.return=TRUE)$ix]
      auxr2     = sapply(candidate, function(jpred) summary(lm(Ypredictand ~ Xpredictor[,c(jpred,selpred)]))$r.squared)
      addpred   = candidate[which(auxr2[ auxr2 > maxr2 ] == max(auxr2[ auxr2 > maxr2 ]) )]
      if(length(addpred) == 0){
        break;
      }
      if(length(addpred) >= 1){
        selpred = c(selpred,addpred)
        Npredf  = length(selpred)
        maxr2   = max(auxr2[ auxr2 > maxr2 ])
      }
    }
    
    Ntime       = length(Ypredictand)
    Xpred       = Xpredictor[,selpred]      # Best set of predictors
    if( Nsyears == 1 ){ # If only one year is left out
      Xfor      = Xpredict[selpred]         # Predictors for year iyear
    }
    if( Nsyears >  1 ){ # If more than 1 year is left out
      Xfor      = Xpredict[,selpred]        # Predictors for year iyear
    }
        
  } # End if
  
  
  # Fit MLR model for year iyear and predict!
  X           = Xpred
  mydf        = data.frame(Ypredictand)
  mydf$X      = X
  bestmodel   = lm(formula = Ypredictand ~ X, mydf)
  Yfit        = bestmodel$fitted.values
  S2          = sum( bestmodel$residuals^2 ) / bestmodel$df.residual
  
  # Compute relative importance among predictors
  if( Npredf == 1){  relimpo = 1  }
  if( Npredf >  1){  relimpo = calc.relimp(bestmodel)$lmg/sum(calc.relimp(bestmodel)$lmg) }
  
  # Case # 1: Only one year dropped out (i.e. leave-one-out cross validation)
  if( Nsyears ==1 ){
    newdata     = data.frame(X = Xfor)
    if ( length(Xfor)>1 ){
      newdata     = data.frame(X = t(Xfor))
      newdata$X   = t(Xfor)
      newdata     = subset( newdata, select = -(1:length(Xfor)))
    }
    zp          = predict(bestmodel,newdata,se.fit=T)
    Yhat        = zp$fit
    #Yse         = zp$se
    Yse         = sqrt( t(c(1,Xfor)) %*% vcov(bestmodel) %*% c(1,Xfor) + S2 )
  }
  
  # Case # 2: More than one year dropped out (i.e. leave-one-out cross validation)
  if( Nsyears > 1 ){
    Yhat        = 1:Nsyears
    Yse         = 1:Nsyears
    for( iyear in 1:Nsyears ){ # Start loop over years
      newdata     = data.frame(X = as.matrix(Xfor)[iyear,])
      if ( length(as.matrix(Xfor)[iyear,])>1 ){
        newdata     = data.frame(X = t(as.matrix(Xfor)[iyear,]))
        newdata$X   = t(as.matrix(Xfor)[iyear,])
        newdata     = subset( newdata, select = -(1:length(as.matrix(Xfor)[iyear,])))
      }
      zp          = predict(bestmodel,newdata,se.fit=T)
      Yhat[iyear] = zp$fit
      #Yse[iyear]  = zp$se
      Yse[iyear]  = sqrt( t(c(1,as.matrix(Xfor)[iyear,])) %*% vcov(bestmodel) %*% c(1,as.matrix(Xfor)[iyear,]) + S2 )
    } # End loop over years
  }
  
  # Generate vector
  if( Npred == 1 ){
    bestcomb='TRUE'
  }
  if( Npred  > 1 ){
    bestcomb = rep(FALSE,Npred)
    bestcomb[selpred] = TRUE
  }
  
  Yreturn     = list(Yhat=Yhat,Yse=Yse,Yfit=Yfit,bestcomb=bestcomb,relimpo=relimpo)
  return(Yreturn)
  
}
